# -*- encoding: binary -*-
#
# To read more about \MogileFS, go to http://mogilefs.org/
#
# Client usage information is available in MogileFS::MogileFS.
module MogileFS

  # Standard error class for most MogileFS-specific errors
  class Error < StandardError; end

  # Raised when a socket remains unreadable for too long.
  class UnreadableSocketError < Error; end

  # Raised when a response is truncated while reading
  # due to network/server # errors)
  class SizeMismatchError < Error; end

  # Raised when checksum verification fails (only while reading deprecated
  # "bigfiles" from the deprecated mogtool(1).
  class ChecksumMismatchError < RuntimeError; end

  # Raised when a backend is in read-only mode
  class ReadOnlyError < Error
    def message # :nodoc:
      'readonly mogilefs'
    end
  end

  # Raised when an upload is impossible
  class EmptyPathError < Error; end

  # Raised when we are given an unsupported protocol to upload to.
  # Currently, the \MogileFS (2.XX) server only supports HTTP and
  # this library is only capable of HTTP.
  class UnsupportedPathError < Error; end

  # Raised when a request (HTTP or tracker) was truncated due to a network or
  # server error.  It may be possible to retry idempotent requests from this.
  class RequestTruncatedError < Error; end

  # Raised when a response from a server (HTTP or tracker) is not in a format
  # that we expected (or truncated..
  class InvalidResponseError < Error; end

  # Raised when all known backends have failed.
  class UnreachableBackendError < Error; end

  # There was an error as a result of the use of the (experimental)
  # pipelining code to the tracker backend
  class PipelineError < Error; end

  # :stopdoc:
  class << self
    # somebody could use IO::Splice from the "io_splice" RubyGem, too
    # don't consider this a stable API, though...
    attr_accessor :io
  end

  # IO.copy_stream was buggy in Ruby 1.9.2 and earlier
  if RUBY_VERSION >= "1.9.3"
    @io = IO
  else
    require "mogilefs/copy_stream"
    @io = MogileFS::CopyStream
  end

  begin
    require 'net/http/persistent'
    NHP = Net::HTTP::Persistent
  rescue LoadError
    require 'mogilefs/nhp_fake'
    NHP = MogileFS::NhpFake
  end

  # autoload rarely-used things:
  autoload :Mysql, 'mogilefs/mysql'
  autoload :Pool, 'mogilefs/pool'
  autoload :Admin, 'mogilefs/admin'
  # :startdoc:
end

require 'mogilefs/util'
require 'mogilefs/socket'
require 'mogilefs/backend'
require 'mogilefs/http_file'
require 'mogilefs/http_reader'
require 'mogilefs/client'
require 'mogilefs/bigfile'
require 'mogilefs/mogilefs'
require 'mogilefs/version' # generated by ./GIT-VERSION-GEN
